/******************************************************************************

 COMP.C

 Compares image to a lookup table.
 
 Subdivides the incoming image into 16 X 16 blocks, and calculates the mean 
 pixel intensity for each block.  The resultant array of means is compared to
 values from a lookup table to determine the number of blocks exceeding the
 threshold values in the lookup table.  A lookup table is selected based on
 the camera, filter, and exposure length, or a set of fixed values may be 
 specified.
 
 Accepts camera images with binning factors of 1, 2, or 4 in both axises.
 
 Writes the result of the comparision into a file for the decision module
 to read, and make a decision on which observation mode to execute next.

 Exits with a return value of -1 if there is any exception, or 0 if all 
 is well.
 
******************************************************************************/
#include <bios.h>
#include <dos.h>
#include <conio.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <fcntl.h>
#include <io.h>
#include <sys\types.h>
#include <sys\stat.h>

#include "at200.h"
#include "idecls.h"
#include "camsys.h"
#include "proto.h"
#include "mode.h"

/* Function prototypes. */

void get_lookup( void );
void show_help( void );
void analyze_image( void );
void show_help ( void );

/* Globals */

/* Actual image data. */

unsigned int img_data[32][512];		/* Raw input image (up to 32 rows)   */

/* 16 X 16 array of mean pixel value per block. */

unsigned int imavg[16][16];		

/* Lookup table. */

unsigned int lkup_val[16][16];

int pixel_max, pixel_min;
unsigned int nval, dispsw;
unsigned int image_pbin, image_sbin, image_rows, image_cols, table;

char filespath[PATHSIZE], lkupflnm[80];
int fd;

/* Get name of last image file that was written. */

/* Read a data lookup table from a file that matches image. */
 
void get_lookup( void )
{
char	lkflname[13], databuff[80];
unsigned int expose_sec, expose_sel, line_no, ii, jj;
FILE 	*fp;

	/* Build file name for table lookup data.
	 * Name is CxFyXzzz.DAT where:
	 * x = camera #, 0 or 1
	 * y = filter #, 0 thru 4,
	 * zzz = exposure duration in seconds.
	 *       This is selected to 1, 10, or 60 second values.
	 */
	
	expose_sec = (unsigned int) ( sel_exp->exposure / 1000L );
	if (expose_sec < 4 )
		expose_sel = 1;
	else if ( expose_sec < 30 )
		expose_sel = 10;
	else 
		expose_sel = 60;
		
	sprintf( lkflname, "C%dF%dX%3d.DAT", 
		sel_cam,
		sel_misc->fw,
		expose_sel );
	for ( ii=0; ii<13; ii++ )		/* spaces to zeroes */
		if ( lkflname[ ii ] == ' ' ) 
			lkflname[ ii ] = '0'; 
	strcpy ( lkupflnm, filespath );		/* Build full pathname */
	strcat ( lkupflnm, lkflname );

	if ( dispsw )
	{
		sprintf ( omsg, "Reading lookup table file: %s\n", lkupflnm );
		disp_omsg( NOERROR );
	}
	if ( ( fp=fopen( lkupflnm, "r" ) ) == NULL )
	{
		sprintf ( omsg, "Can't open lookup table file %s\n", lkupflnm );
		disp_omsg( ERROR );
		return;
	}
	for ( line_no=0, ii=0; ii<16; ii++ )
	{
	   fgets( databuff, 80, fp );
	   line_no++;
	   jj = sscanf( &databuff[10], "%d %d %d %d %d %d %d %d",
		   &lkup_val[ii][0], &lkup_val[ii][1],
		   &lkup_val[ii][2],
		   &lkup_val[ii][3], &lkup_val[ii][4],
		   &lkup_val[ii][5],
		   &lkup_val[ii][6], &lkup_val[ii][7] );
	   if ( jj != 8 )
	   {
	   	sprintf( omsg, "ERROR: Invalid value in threshold data file at line %d.\n",
	   		line_no );
	   	disp_omsg( ERROR );
		fclose ( fp );
	   	return;
	   }
	   fgets( databuff, 80, fp );
	   line_no++;
	   jj = sscanf( &databuff[10], "%d %d %d %d %d %d %d %d",
		   &lkup_val[ii][8], &lkup_val[ii][9], &lkup_val[ii][10],
		   &lkup_val[ii][11], &lkup_val[ii][12], &lkup_val[ii][13],
		   &lkup_val[ii][14], &lkup_val[ii][15] );
	   if ( jj != 8 )
	   {
	   	sprintf( omsg, "ERROR: Invalid value in threshold data file at line %d.\n",
	   		line_no );
	   	disp_omsg( ERROR );
		fclose ( fp );
	   	return;
	   }
	}
	fclose ( fp );
}

/* Read a camera image file, build 16 X 16 table of average pixel
 * intensity levels for comparasion to lookup table.
 */ 

void analyze_image(void)
{
register unsigned row, col;
unsigned int	ii, jj, y, x, z, num, blkx, blky;
unsigned long	temp;

	/* Clear imavg table */
	
	for ( y=0; y<16; y++ )
		for ( x=0; x<16; x++ )
			imavg[y][x] = 0;

	/* Check if image to be processed is at an acceptable binning factor. */


	if ( !(   ( ( image_pbin == 1 ) && ( image_sbin == 1 ) )
	       || ( ( image_pbin == 2 ) && ( image_sbin == 2 ) )
	       || ( ( image_pbin == 4 ) && ( image_sbin == 4 ) ) ) )
	{
		sprintf( omsg, "ERROR: binning must be 1X1, 2X2, or 4X4 for image.\n");
		disp_omsg( ERROR );
	}	      
	
	y = image_rows / image_pbin;	/* y = actual # of pixels in y axis */
	x = image_cols / image_sbin;	/* x = actual # of pixels in y axis */
	blky = y / 16;			/* blky = y pixels down   in block */
	blkx = x / 16;			/* blkx = x pixels across in block */
	num = blkx * blky;		/* number of pixels in block. */
	
	/* Process the input image data */

	for( row=0; row<y; row++ )       /* Each row of image data */
	{
		z = row % blky;
		read( fd, img_data[ z ], x*2 ); /* Read one row of data */
		if ( z == ( blky - 1 ) )	/* last row for this block ? */
		{
			/* Process this group of blocks */
			
			for( temp = 0L, col=0; col<x; col += blkx )
			{
			   for ( ii=0; ii<blky; ii++ ) /* blky pixels down */ 
			      for ( jj=0; jj<blkx; jj++ ) /* and blkx across */
			         temp += img_data[ii][col + jj];
			   temp += num / 2;
			   imavg[ (row / blky) ][ (col / blkx) ] = 
			   	(unsigned int) ( temp / num );
			   temp = 0L;
			}
		}
	}
}

/* KI compare command. */

pv_ptr compare ( register pv_ptr p )
{
unsigned int ii, jj, tval, x, y, line_no, count, decision;
FILE *fp;
char	databuff[82];

char	*token;
char	cmdargs[82];

	table = dispsw = 0;
	nval = 128;		/* 50% of 16 X 16 blocks */
	cmdargs[0] = '\0';

	/* See if there is a command line argument string. */
	
	if ( (**p).type == LITERAL )
	{
		strcpy ( cmdargs, (**p).name );
		++p;
		set_cmdsrc ( cmdargs, CMD_LINE );	/* tokenize from here */
		while ( ( token = get_nxt_tok() ) != (void *) NULL )
		{
		    if( token[0] == '-' )
    			switch( token[1] ) 
    			{

      			case 'e': case 'E':		/* Do nothing */
				return p;		/* *** TEMPORARY */
      	 
			/* Display output switch. */

      			case 'd': case 'D':
      	  			dispsw = 1;
      	  			break;
      	 
			/* Number of values that need to exceed threshold. */

      			case 'n': case 'N':
      	  			nval = get_nxt_number();
      	  			break;
      	 
			/* Load lookup table with uniform value given by '-t nnn' arg. */

			case 't': case 'T':
				token = get_nxt_tok();
      	  			tval = atoi ( token );
      	  			sprintf ( lkupflnm, "-t %s", token );
      	  			for ( ii=0; ii<16; ii++ )
      	     		 	   for ( jj=0; jj<16; jj++ )
					lkup_val[ii][jj] = tval;
	  			table = 1;    /* Show that lookup table is loaded. */
      	  			break;
      	  
			/* Use lookup table specified by '-l xxx" arg. */
			/* (overrides normal lookup table selection) */

        		case 'l': case 'L':
        		    strcpy ( lkupflnm, get_nxt_tok() );
			    if ( ( fp=fopen( lkupflnm, "r" ) ) == NULL )
			    {
				sprintf ( omsg, "Can't open %s.\n", 
					lkupflnm );
				disp_omsg( ERROR );
				break;
			    }
			    for ( line_no=0, ii=0; ii<16; ii++ )
			    {
			   	fgets( databuff, 80, fp );
			   	line_no++;
			   	jj = sscanf( &databuff[10], "%d %d %d %d %d %d %d %d",
				   &lkup_val[ii][0], &lkup_val[ii][1],
				   &lkup_val[ii][2],
				   &lkup_val[ii][3], &lkup_val[ii][4],
				   &lkup_val[ii][5],
				   &lkup_val[ii][6], &lkup_val[ii][7] );
		   		if ( jj != 8 )
		   		{
		   			printf("ERROR: Invalid value in threshold data file at line %d.\n",
		   				line_no );
		   		}
		   		fgets( databuff, 80, fp );
		   		line_no++;
		   		jj = sscanf( &databuff[10], "%d %d %d %d %d %d %d %d",
			  	&lkup_val[ii][8], &lkup_val[ii][9], &lkup_val[ii][10],
			  	 &lkup_val[ii][11], &lkup_val[ii][12], &lkup_val[ii][13],
			  	 &lkup_val[ii][14], &lkup_val[ii][15] );
		   		if ( jj != 8 )
		   		{
		   			printf("ERROR: Invalid value in threshold data file at line %d.\n",
		   				line_no );
				}
			    }
			    fclose ( fp );
			    table = 1;
			    break;

        		default:
				show_help();
        			break;
			}
		}
	}

  	get_lookup();   /* Load lookup table. */
	
	analyze_image();		/* Do it. */

	/* Scan result table, compare to lookup table. */

	count = 0;
	if ( dispsw )
	{
		sprintf( omsg, "Results:\n\n" );
		disp_omsg( NOERROR );

		sprintf( omsg, "------|" );
		for ( ii=0; ii<10; ii++ )
			sprintf( omsg, "%s    %2d|", omsg, ii );
	}
	
	for ( y=0; y<16; y++ )
	{
		if ( dispsw )
		{
			sprintf( omsg, "%s\n%2d    |", omsg, y );
		}
		for ( x=0; x<16; x++ )
		{
			if ( dispsw )
			{
				sprintf( omsg, "%s %5d ", omsg, imavg[y][x] );
				if ( !((x+1) % 10) )
					sprintf( omsg, "%s\n      |", omsg );
			}
			if ( imavg[y][x] > lkup_val[y][x] )
				count++;
		}
		if ( dispsw )
		{
			sprintf( omsg, "%s\n", omsg );
			disp_omsg( NOERROR );
		}
	}
	
	/* Set decision flag based on whether number of blocks exceeding
	 * lookup table threshold values exceeded 'nval'.
	 */

	if ( count > nval )
		decision = 1;
	else
		decision = 0;

	if ( dispsw )
	{
		sprintf ( omsg, "\nCAMERA %d, COUNT %3d, NVAL %3d, DECISION %d\n",
			sel_cam,
			count, nval, decision, decision );
		disp_omsg( NOERROR );
		sprintf ( omsg, "          LOOKUP FILE: %s\n", lkupflnm );
		disp_omsg( NOERROR );
	}

	return p;
}

void show_help ( void )
{
  sprintf( omsg, "USAGE: COMPARE [-t nnn][-d][-n mmm][-l filename]\n"
         "  Where:\n"
         "    -t  nnn       Use all 'nnn' values for table lookup data\n"
         "    -d            Display results\n"
         "    -n  mmm       Number of values that need to exceed threshold\n"
         "    -l 'filename' Name of table lookup file to use. (override)\n" );
  disp_omsg( NOERROR );
}
